"""
Classes for rendering 3D images of molecules and densities, primarily with blender and beautiful atoms.

Also see vmd.py for an older render engine.
"""
from pathlib import Path
import os
import copy
import subprocess
import yaml
from PIL import Image
import math
import numpy
import warnings
import json

import digichem.log
from digichem.exception.base import File_maker_exception
from digichem.file.base import File_converter
from digichem.image.base import Cropable_mixin
from digichem.datas import get_resource


class Render_maker(File_converter, Cropable_mixin):
    """
    ABC for classes that make 3D renders from cube files.
    """
        
    # Text description of our input file type, used for error messages etc. This can be changed by inheriting classes.
    input_file_type = "cube"
    # Text description of our output file type, used for error messages etc. This can be changed by inheriting classes.
    output_file_type = "render"
    
    def __init__(
            self,
            *args,
            cube_file = None,
            rotations = None,
            auto_crop = True,
            resolution = 1024,
            also_make_png = True,
            isovalue = 0.2,
            num_cpu = 1,
            **kwargs):
        """
        Constructor for Image_maker objects.
        
        :param output: The path to write image files to. As more than one image can be created by this class, this name is used as the base name and is modified for each image. The suffix will be honoured (rendered imgaes will use it as a hint for their output format).
        :param cube_file: The path to a cube_file to use to render images.
        :param rotations: A list of tuples of rotations, where the first index in the tuple specifies the axis to rotate about and the second is the angle to rotate (in radians).
        :param auto_crop: If False, images will not have excess white space cropped.
        :param resolution: The max width or height of the rendered images in pixels.
        :param also_make_png: If True, additional images will be rendered in PNG format. This option is useful to generate higher quality images alongside more portable formats.
        :param isovalue: The isovalue to use for rendering isosurfaces. Has no effect when rendering only atoms.
        :param num_cpu: The number of CPUs for multithreading.
        :param blender_executable:
        """
        super().__init__(*args, input_file = cube_file, **kwargs)
        # Save our translations list.
        #self.translations = translations if translations is not None else (0,0,0)
        # This is not used, why not?
        self.translations = (0,0,0)
        # Save our rotations list.
        self._rotations = rotations if rotations is not None else []
        
        # Some options that control how we function.
        self.auto_crop = auto_crop
        self.target_resolution = resolution
        self.also_make_png = also_make_png
        self.isovalue = isovalue
        self.num_cpu = num_cpu
        
        # TODO: These.
        self.primary_colour = "red"
        self.secondary_colour = "blue"
                
        # These 4 attributes are file paths to the four images we create.
        # We'll keep the same file extension as the was given to us.
        self.file_path = {
            'x0y0z0': self.output.with_suffix(".x0y0z0" + self.output.suffix),
            'x90y0z0': self.output.with_suffix(".x90y0z0" + self.output.suffix),
            'x0y90z0': self.output.with_suffix(".x0y90z0" + self.output.suffix),
            'x45y45z45': self.output.with_suffix(".x45y45z45"  + self.output.suffix),
            # TODO: This should respect also_make_png, but works fine for now (these are actually unused).
            # There are higher quality PNG versions.
             'x0y0z0_big': self.output.with_suffix(".x0y0z0.png"),
             'x90y0z0_big': self.output.with_suffix(".x90y0z0.png"),
             'x0y90z0_big': self.output.with_suffix(".x0y90z0.png"),
             'x45y45z45_big': self.output.with_suffix(".x45y45z45.png")
            }
        
    def get_image(self, name = 'file'):
        """
        Get the path to one of the images that this class represents, rendering the image to file first if necessary.
        
        The functioning of this method is controlled by the dont_modify & use_existing flags.
        
        You can also use the normal python attribute mechanism (either through getattr() or dot notation) to get these paths.
        
        :raises KeyError: If name is not the name of one of the images this class represents.
        :param name: The name of an image to get. Depends on the images in self.file_path.
        :return: A pathlib Path object pointing to the image represented by name, or None if no image could be created.
        """
        return self.safe_get_file(name)
    
    @property
    def rotations(self):
        # Digichem rotates the wrong way round for some reason, reverse for our rendering engine.
        return [(axis, theta) for axis, theta in self._rotations]


class Batoms_renderer(Render_maker):
    """
    Class for rendering images with Beautiful Atoms and Blender.
    """
        
    # Text description of our input file type, used for error messages etc. This can be changed by inheriting classes.
    input_file_type = "cube"
    # Text description of our output file type, used for error messages etc. This can be changed by inheriting classes.
    output_file_type = "render"
    
    # TODO: Do we need this? Can probably remove if/when VMD is dropped.
    # Name of the section where we get some specific configs.
    options_name = "orbital"
    
    test_resolution = 300
    test_samples = 5
    
    @property
    def batoms_script_path(self):
        """
        Get the file path to the VMD script to be used by this class to render images (as a pathlib Path).
        """
        return get_resource('data/batoms/batoms-renderer.py')
    
    def __init__(
            self,
            *args,
            cube_file = None,
            rotations = None,
            auto_crop = True,
            resolution = 1024,
            render_samples = 32,
            stack = 3,
            also_make_png = True,
            isovalue = 0.02,
            blender_executable = None,
            cpus = 1,
            num_cpu = 1,
            perspective = "perspective",
            logging = False,
            **kwargs):
        """
        Constructor for Image_maker objects.
        
        :param output: The path to write image files to. As more than one image can be created by this class, this name is used as the base name and is modified for each image. The suffix will be honoured (rendered imgaes will use it as a hint for their output format).
        :param cube_file: The path to a cube_file to use to render images.
        :param rotations: A list of tuples of rotations, where the first index in the tuple specifies the axis to rotate about and the second is the angle to rotate (in radians).
        :param auto_crop: If False, images will not have excess white space cropped.
        :param resolution: The max width or height of the rendered images in pixels.
        :param render_samples: The number of render samples, more results in longer render times but higher quality image.
        :param stack: The number of copies of the image to composite together to avoid transparency artifacts.
        :param also_make_png: If True, additional images will be rendered in PNG format. This option is useful to generate higher quality images alongside more portable formats.
        :param isovalue: The isovalue to use for rendering isosurfaces. Has no effect when rendering only atoms.
        :param blender_executable: Bath to the blender executable (can be None to use a default).
        :param cpus: DEPREACTED: Number of parallel threads to render with (use num_cpu instead)
        :param num_cpu: Number of parallel threads to render with.
        :param perspective: Perspective mode (orthographic or perspective)
        """
        if cpus != 1:
            warnings.warn("cpus is deprecated, use num_cpu instead", DeprecationWarning)
        
        if num_cpu == 1 and cpus != 1:
            num_cpu = cpus
        
        super().__init__(
            *args,
            cube_file = cube_file,
            rotations = rotations,
            auto_crop = auto_crop,
            resolution = resolution,
            also_make_png = also_make_png,
            isovalue = math.fabs(isovalue),
            num_cpu = num_cpu,
            **kwargs
        )
        
        # Blender specific options.
        self.render_samples = render_samples
        self.perspective = perspective
        self.stack = stack

        self.logging = logging
        
        # Use explicit blender location if given.
        if blender_executable is not None:
            self.blender_executable = blender_executable
        
        else:
            # Otherwise, use a default location.
            self.blender_executable = get_resource('data/batoms/blender/blender')
            
    @classmethod
    def from_options(self, output, *, cube_file = None, rotations = None, cpus = None, num_cpu = 1, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            cube_file = cube_file,
            rotations = rotations,
            auto_crop = options['render']['auto_crop'],
            resolution = options['render']['resolution'],
            render_samples = options['render']['batoms']['render_samples'],
            stack = options['render']['batoms']['stacking'],
            isovalue = options['render'][self.options_name]['isovalue'],
            use_existing = options['render']['use_existing'],
            dont_modify = not options['render']['enable_rendering'],
            blender_executable = options['render']['batoms']['blender'],
            # Deprecated...
            cpus = cpus if cpus is not None else options['render']['batoms']['cpus'],
            num_cpu = num_cpu,
            perspective = options['render']['batoms']['perspective'],
            logging = options['logging']['render_logging'],
            **kwargs
        )
        
    def blender_signature(self, *targets, padding = 1.0):
        """
        The signature passed to subprocess.run used to call Blender. Inheriting classes should write their own implementation.
        """
        args = [
            # Blender args.
            f"{self.blender_executable}",
            # Run in background.
            "-b",
            # Our script.
            "-P", f"{self.batoms_script_path}",
            "--",
            # Script specific args.
            f"{self.input_file}",
            # f"{output}",
            # Keywords.
            "--cpus", f"{self.num_cpu}",
            # "--orientation", "{}".format(orientation[0]), "{}".format(orientation[1]), "{}".format(orientation[2]),
            # "--resolution", f"{resolution}",
            # "--render-samples", f"{samples}",
            "--perspective", f"{self.perspective}",
            "--padding", f"{padding}",
        ]
        for orientation, resolution, samples, mini_file_name in targets:
            args.extend([
                "--multi",
                "{}".format(orientation[0]), "{}".format(orientation[1]), "{}".format(orientation[2]),
                f"{resolution}",
                f"{samples}",
                mini_file_name
            ])

        if len(self.rotations) > 0:
            args.append("--rotations")

            # Add rotations.
            for rotation in self.rotations:
                args.append(json.dumps(rotation))
        
        return args
    
    #def run_blender(self, output, resolution, samples, orientation):

    def run_blender(self, *targets):
        """
        Render a (number of) images with blender.

        :param samples: How many render samples to use.
        :param *targets: Images to render, each is a tuple if (orientation, resolution, samples, path).
        """
        # Render with batoms.
        env = dict(os.environ)
        
        # Disabling the user dir helps prevent conflicting installs of certain packages
        env["PYTHONNOUSERSITE"] = "1"
        #env["PYTHONPATH"] = ":" + env["PYTHONPATH"]
        blender_process = None
        
        # Run Blender, which renders our image for us.
        try:
            blender_process = subprocess.run(
                self.blender_signature(*targets),
                stdin = subprocess.DEVNULL,
                stdout = subprocess.PIPE if not self.logging else None,
                stderr = subprocess.STDOUT,
                universal_newlines = True,
                check = True,
                env = env,
            )
            
            for target in targets:
                if not target[3].exists():
                    raise File_maker_exception(self, "Could not find render file '{}'".format(target[3]))

        except FileNotFoundError:
            raise File_maker_exception(self, "Could not locate blender executable '{}'".format(self.blender_executable))
        
        except subprocess.CalledProcessError as e:
            if not self.logging:
                digichem.log.get_logger().error("Blender did not exit successfully, dumping output:\n{}".format(e.stdout))

        except Exception:
            if not self.logging and blender_process is not None:
                digichem.log.get_logger().error("Blender did not exit successfully, dumping output:\n{}".format(blender_process.stdout))
            
            raise
        
    
    def make_files(self):
        """
        Make the image files referenced by this object.
        
        The new image will be written to file.
        """
        # TODO: This mechanism is clunky and inefficient if only one image is needed because its based off the old VMD renderer. With batoms we can do much better.
        angles = {
            "x0y0z0": [
                (0,0,0),
                self.test_resolution if self.auto_crop else self.target_resolution,
                self.test_samples if self.auto_crop else self.render_samples,
                self.file_path['x0y0z0'].with_suffix(".tmp.png")
            ],
            "x90y0z0": [
                (1.5708, 0, 0),
                self.test_resolution if self.auto_crop else self.target_resolution,
                self.test_samples if self.auto_crop else self.render_samples,
                self.file_path['x90y0z0'].with_suffix(".tmp.png")
            ],
            "x0y90z0": [
                (0, 1.5708, 0),
                self.test_resolution if self.auto_crop else self.target_resolution,
                self.test_samples if self.auto_crop else self.render_samples,
                self.file_path['x0y90z0'].with_suffix(".tmp.png")
            ],
            "x45y45z45": [
                (0.785398, 0.785398, 0.785398),
                self.test_resolution if self.auto_crop else self.target_resolution,
                self.test_samples if self.auto_crop else self.render_samples,
                self.file_path['x45y45z45'].with_suffix(".tmp.png")
            ],
        }
        
        try:
            # First we'll render a test image at a lower resolution. We'll then crop it, and use the decrease in final resolution to know how much bigger we need to render in our final image to hit our target resolution.
            # Unless of course auto_crop is False, in which case we use our target resolution immediately.
            self.run_blender(*list(angles.values()))
            
            if self.auto_crop:
                # Load the test image and autocrop it.
                for angle, target in angles.items():
                    with Image.open(target[3], "r") as test_im:
                        small_test_im = self.auto_crop_image(test_im)
                        
                    # Get the cropped size. We're interested in the largest dimension, as this is what we'll output as.
                    cropped_resolution = max(small_test_im.size)
                    
                    # From this we can work out the ratio between our true resolution and the resolution we've been asked for.
                    resolution_ratio = cropped_resolution / self.test_resolution

                    # Update the target properties.
                    angles[angle] = [
                        # Orientation is the same.
                        angles[angle][0],
                        # New resolution.
                        int(self.target_resolution / resolution_ratio),
                        # New samples.
                        self.render_samples,
                        # New filename,
                        angles[angle][3]
                    ]
                    
                self.run_blender(*list(angles.values()))
            
        except Exception:
            raise File_maker_exception(self, "Error in blender rendering")
        
        # Convert to a better set of formats.
        # Open the files we just rendered.
        for angle, target in angles.items():
            with Image.open(target[3], "r") as im:
                
                # If we've been asked to autocrop, do so.
                if self.auto_crop:
                    try:
                        cropped_image = self.auto_crop_image(im)
                    except Exception:
                        raise File_maker_exception(self, "Error in post-rendering auto-crop")
                else:
                    cropped_image = im

                # Transparency stacking.
                # This 'hack' takes several copies of the same transparent image and layers them atop each other.
                # This avoids problems with isosurfaces being too transparent with respect to the background (and
                # thus loosing definition).
                stacked = cropped_image.copy()

                for _ in range(self.stack):
                    stacked.alpha_composite(cropped_image)

                cropped_image = stacked
                
                # Save as a higher quality png if we've been asked to.
                if self.also_make_png:
                    cropped_image.save(self.file_path[angle + "_big"])
                    
                # Remove transparency, which isn't supported by JPEG (which is essentially the only format we write here).
                # TODO: Check if the output format can support transparency or not.
                new_image = Image.new("RGBA", cropped_image.size, "WHITE") # Create a white rgba background
                new_image.paste(cropped_image, (0, 0), cropped_image)              # Paste the image on the background. Go to the links given below for details.
                cropped_image = new_image.convert('RGB')
                
                # Now save in our main output format.
                cropped_image.save(self.file_path[angle])
                
                # And delete the old .png.
                os.remove(target[3])


class Structure_image_maker(Batoms_renderer):
    """
    Class for creating structure images.
    """
        
        
class Orbital_image_maker(Structure_image_maker):
    """
    Class for creating orbital images.
    """
    
    def blender_signature(self, output, resolution, samples, orientation, isotype = "both", isocolor = "sign"):
        """
        The signature passed to subprocess.run used to call Blender. Inheriting classes should write their own implementation.
        """
        sig = super().blender_signature(output, resolution, samples, orientation)
        sig.extend(["--isovalues", f"{self.isovalue}"])
        sig.extend(["--isotype", isotype])
        sig.extend(["--isocolor", isocolor])
        
        return sig
    

class Difference_density_image_maker(Orbital_image_maker):
    
    # Name of the section where we get some specific configs.
    options_name = "difference_density"


class Spin_density_image_maker(Orbital_image_maker):
    """
    Class for creating spin density images.
    """
    
    # Name of the section where we get some specific configs.
    options_name = "spin"
    
    def __init__(self, *args, spin = "both", **kwargs):
        """
        Constructor for Spin_density_image_maker objects.
        
        See Orbital_image_maker for a complete constructor.
        :param spin: A string indicating which net-spins to render, either 'positive', 'negative' or 'both'.
        """
        super().__init__(*args, **kwargs)
        self.spin = spin
                
    def blender_signature(self, output, resolution, samples, orientation):
        """
        The signature passed to subprocess.run used to call Blender. Inheriting classes should write their own implementation.
        """
        sig = super().blender_signature(output, resolution, samples, orientation, isotype = self.spin)
        return sig


class Alpha_orbital_image_maker(Orbital_image_maker):
    pass


class Beta_orbital_image_maker(Orbital_image_maker):
    pass


class Density_image_maker(Orbital_image_maker):
    """
    Class for creating spin density images.
    """
    
    # Name of the section where we get some specific configs.
    options_name = "density"
    
    @property
    def type(self):
        """
        The density type.
        """
        return self.input_file.type
    
    def blender_signature(self, output, resolution, samples, orientation):
        """
        The signature passed to subprocess.run used to call Blender. Inheriting classes should write their own implementation.
        """
        # There is no negative density for total density.
        sig = super().blender_signature(output, resolution, samples, orientation, isotype = "positive")
        return sig


class Combined_orbital_image_maker(Orbital_image_maker):
    """
    Class for creating images with both the HOMO and LUMO shown together.
    """
    
    vmd_script ="generate_combined_orbital_images.tcl"
    
    def __init__(self, *args, cube_file = None, LUMO_cube_file = None, **kwargs):
        """
        Constructor for combined orbital image maker objects.
        
        :param output: Path to write to. See the constructor for Batoms_renderer for how this works.
        :param HOMO_cube_file: Path to the HOMO cube file.
        :param LUMO_cube_file: Path to the LUMO cube file.
        :param *args: See the constructor for Batoms_renderer for further options.
        :param **kwargs: See the constructor for VMD_image_maker for further options.
        """
        super().__init__(*args, cube_file = cube_file, **kwargs)
        self.LUMO_cube_file = LUMO_cube_file
        
    @property
    def HOMO_cube_file(self):
        return self.input_file
        
    @classmethod
    def from_options(self, output, *, HOMO_cube_file = None, LUMO_cube_file = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return super().from_options(
            output,
            cube_file = HOMO_cube_file,
            LUMO_cube_file = LUMO_cube_file,
            options = options,
            **kwargs
        ) 

    def check_can_make(self):
        """
        Check whether it is feasible to try and render the image(s) that we represent.
        
        Reasons for rendering not being possible are varied and are up to the inheriting class, but include eg, a required input (cube, fchk) file not being given.
        
        This method returns nothing, but will raise a File_maker_exception exception if the rendering is not possible.
        """
        try:
            if self.HOMO_cube_file is None or self.HOMO_cube_file.safe_get_file() is None:
                raise File_maker_exception(self, "No HOMO cube file is available")
        except AttributeError:
            if not hasattr(self.HOMO_cube_file, 'safe_get_file'):
            # Input file does not have a safe_get_file() method.
                pass

            else:
                raise

        try:
            if self.LUMO_cube_file is None  or self.LUMO_cube_file.safe_get_file() is None:
                raise File_maker_exception(self, "No LUMO cube file is available")
        
        except AttributeError:
            if not hasattr(self.LUMO_cube_file, 'safe_get_file'):
            # Input file does not have a safe_get_file() method.
                pass

            else:
                raise
    
    def blender_signature(self, output, resolution, samples, orientation):
        """
        The signature passed to subprocess.run used to call Blender. Inheriting classes should write their own implementation.
        """
        sig = super().blender_signature(output, resolution, samples, orientation, isotype = "both", isocolor = "cube")
        sig.extend(["--second_cube", f"{self.LUMO_cube_file}"])
        return sig


class Dipole_image_maker(Structure_image_maker):
    """
    Class for creating dipole images.
    """
    
    
    def __init__(self, *args, dipole_moment = None, magnetic_dipole_moment = None, scaling = 1, magnetic_scaling = 1, **kwargs):
        """
        Constructor for Dipole_image_maker objects.
        
        :param output: Path to write to. See the constructor for VMD_image_maker for how this works.
        :param cube_file: A Gaussian cube file to use to render the new images.
        :param dipole_moment: A Dipole_moment object that will be rendered as a red arrow in the scene.
        :param magnetic_dipole_moment: A second dipole moment object that will be rendered as a blue arrow in the scene.
        :param scaling: A factor to scale the dipole moment by.
        :param magnetic_scaling: A factor to scale the magnetic dipole moment by.
        :param *args: See the constructor for VMD_image_maker for further options.
        :param **kwargs: See the constructor for VMD_image_maker for further options.
        """
        super().__init__(*args, **kwargs)
        self.dipole_moment = dipole_moment
        self.magnetic_dipole_moment = magnetic_dipole_moment
        self.scaling = scaling
        self.magnetic_scaling = magnetic_scaling
        self.electric_arrow_colour = "red"
        self.magnetic_arrow_colour = "green"
        
    def get_dipoles(self):
        """
        Get required dipole information as a nested list.
        """
        dipoles = []
        # First electric (if we have it).
        for dipole, scaling, colour in [
            (self.dipole_moment, self.scaling, [1.0, 0.0, 0.0, 1.0]),
            (self.magnetic_dipole_moment, self.magnetic_scaling, [0.0, 1.0, 0.0, 1.0]),
        ]:
            if dipole is not None:
                dipoles.append([
                    [float(coord * scaling) for coord in dipole.origin_coords],
                    [float(coord * scaling) for coord in dipole.vector_coords],
                    colour
                ])
        
        return dipoles
    
    def blender_signature(self, output, resolution, samples, orientation):
        """
        The signature passed to subprocess.run used to call Blender. Inheriting classes should write their own implementation.
        """
        sig = super().blender_signature(output, resolution, samples, orientation)
        sig.append("--dipoles")
        for dipole in self.get_dipoles():
            #sig.append(yaml.safe_dump(dipole))
            sig.append(json.dumps(dipole))
        sig.extend(["--alpha", "0.75"])
        return sig

    
class Permanent_dipole_image_maker(Dipole_image_maker):
    """
    Class for creating dipole images.
    """
        
    @classmethod
    def from_options(self, output, *, dipole_moment = None, cube_file = None, rotations = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """
        return super().from_options(
            output,
            dipole_moment = dipole_moment, 
            cube_file = cube_file,
            rotations = rotations,
            scaling = options['render']['dipole_moment']['scaling'],
            options = options,
            **kwargs
        )
    
    def check_can_make(self):
        """
        Check whether it is feasible to try and render the image(s) that we represent.
        
        Reasons for rendering not being possible are varied and are up to the inheriting class, but include eg, a required input (cube, fchk) file not being given.
        
        This method returns nothing, but will raise a File_maker_exception exception if the rendering is not possible.
        """ 
        super().check_can_make()
        
        # Also make sure we have a dipole.
        if self.dipole_moment is None:
            raise File_maker_exception(self, "No dipole moment is available.")
    

class Transition_dipole_image_maker(Dipole_image_maker):
    """
    Class for creating TDM images.
    """
    
    def check_can_make(self):
        """
        Check whether it is feasible to try and render the image(s) that we represent.
        
        Reasons for rendering not being possible are varied and are up to the inheriting class, but include eg, a required input (cube, fchk) file not being given.
        
        This method returns nothing, but will raise a File_maker_exception exception if the rendering is not possible.
        """ 
        super().check_can_make()
        
        # Also make sure we have a dipole.
        if self.dipole_moment is None and self.magnetic_dipole_moment is None:
            raise File_maker_exception(self, "No dipole moment is available.")
        
    @classmethod
    def from_options(self, output, *, dipole_moment = None, cube_file = None, magnetic_dipole_moment, rotations = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """
        return super().from_options(
            output,
            dipole_moment = dipole_moment,
            magnetic_dipole_moment = magnetic_dipole_moment,
            cube_file = cube_file,
            rotations = rotations,
            scaling = options['render']['dipole_moment']['scaling'],
            magnetic_scaling = options['render']['transition_dipole_moment']['magnetic_scaling'],
            options = options,
            **kwargs
        )   



import hashlib
import json
from uuid import uuid4

from cclib.bridge.cclib2pyscf import cclibfrommethods

from digichem.parse.base import Parser_abc
import digichem.log

class Pyscf_parser(Parser_abc):
    """
    Top level class for parsing output from pyscf data.
    """
    
    def __init__(self, mol_name, methods, **kwargs):
        self.methods = methods
        self.mol_name = mol_name
        super().__init__(**kwargs)

    def _parse(self):
        self.data = cclibfrommethods(**self.methods)
        
        try:
            # Try to generate a checksum from metadata.
            self.data._id = hashlib.sha1(json.dumps(self.data.metadata, sort_keys = True).encode('utf-8')).hexdigest()

        except Exception:
            # No luck, something in metadata must be unhashable.
            digichem.log.get_logger().error("Unable to generate hash ID from calculation metadata, using random ID instead", exc_info = True)
            # TODO: Think of a better way to do this.
            self.data._id = hashlib.sha1(uuid4().hex.encode('utf-8')).hexdigest()
        
        self.data.metadata['name'] = self.mol_name
        self.data._aux = {'methods': self.methods}
            


"""Tests for handling basis sets from BSE."""
import pytest

from digichem.basis import BSE_basis_set

def test_basis():
    """Can we load a simple basis set?"""
    basis = BSE_basis_set({
        "STO-3G":"all"
    })
    basis.to_format()

def test_ecps():
    """Can we load a simple basis set eith ecps?"""
    basis = BSE_basis_set({
        "LANL2DZ":"all"
    })
    assert basis.has_ECPs()

@pytest.mark.parametrize("set", [
    {"STO-3G":"all"},
    {"STO-3G":"C"},
    {"STO-3G":"C,H"},
    {"STO-3G":"H","def2-SVP":"C"}
], ids = ["all", "C", "C,H", "mixed"])
def test_meta(set):
    """Can we handle basis set metadata correctly?"""
    basis = BSE_basis_set(set)
    str(basis)

@pytest.mark.parametrize("filter, number", [
    ("C", 1),
    ("C,H", 2),
    (("C", "H"), 2)
], ids = ["C", "C,H", list])
def test_filter(filter, number):
    """Can we return only parts of a basis set?"""
    basis = BSE_basis_set({
        "STO-3G":"all"
    })
    output = basis.to_format("gaussian94", filter)
    assert output.count("****") == number

def test_mixed_basis():
    """Can we combine multiple basis sets?"""
    basis = BSE_basis_set({
        "STO-3G":"H",
        "def2-SVP" : "C"
    })

    basis.to_format("gaussian94")

def test_mixed_basis_filter():
    """Can we combine multiple basis sets?"""
    basis = BSE_basis_set({
        "STO-3G":"H",
        "def2-SVP" : "C"
    })

    output = basis.to_format("gaussian94", "C,H")
    assert output.count("****") == 2

def test_wrong_basis():
    """Check a non-existent basis set is flagged"""
    basis = BSE_basis_set({
        "STO-3G":"H",
        "defs2-SVP" : "C"
    })

    with pytest.raises(KeyError):
        basis.to_format("gaussian94")


from pathlib import Path
import subprocess
import math
from PIL import Image
from uuid import uuid4
import os
from math import fabs
import shutil

from digichem.exception.base import File_maker_exception
from digichem.image.render import Render_maker
import digichem.log
from digichem.datas import get_resource


class VMD_image_maker(Render_maker):
    """
    Class for generating image files from Gaussian outputs using VMD.
    
    Nearly all of the work here is done by other programs, primarily VMD (https://www.ks.uiuc.edu/Research/vmd/). These classes are mostly just wrappers.
    """
    
    # The name of the vmd/tcl script (relative to the vmd script folder) used to render images. Inheriting classes should set this to an appropriate script file.
    vmd_script = ""
    # The filename extension to use for molecule scene files produced by vmd.
    scene_file_extension = ".scene"
    # 'Path' to the vmd executable.
    #vmd_executable = "vmd"
    # 'Path' to the tachyon executable.
    #tachyon_executable = "tachyon"
    # The initial resolution at which an image is rendered. This test image will then be discarded once relevant info has been extracted.
    test_resolution = 300
    
    # Name of the section where we get some specific configs.
    options_name = "orbital"
    
    def __init__(self, *args, cube_file = None, rotations = None, auto_crop = True, rendering_style = "pastel", resolution = 1024, also_make_png = True, isovalue = 0.2, 
                 vmd_executable = "vmd", tachyon_executable = "tachyon", vmd_logging = False, num_cpu = 1,
                 **kwargs):
        """
        Constructor for Image_maker objects.
        
        :param output: The path to write image files to. As more than one image can be created by this class, this name is used as the base name and is modified for each image. The suffix will be honoured (rendered imges will use it as a hint for their output format.
        :param cube_file: The path to a cube_file to use to render images.
        :param rotations: A list of tuples of rotations, where the first index in the tuple specifies the axis to rotate about and the second is the angle to rotate (in radians).
        :param auto_crop: If False, images will not have excess white space cropped.
        :param rendering_style: A string describing the rendering style to use, either 'digichem' or 'gaussian'.
        :param resolution: The max width or height of the rendered images in pixels.
        :param also_make_png: If True, additional images will be rendered in PNG format. This option is useful to generate higher quality images alongside more portable formats. If 'output' is a .png file, then it is wise to set this option to False (otherwise two png files will be rendered, which is a waste).
        :param isovalue: The isovalue to use for rendering isosurfaces. Has no effect when rendering only atoms.
        :param num_cpu: Number of CPUs to use for multithreading.
        :param vmd_executable: 'Path' to the vmd executable to use for image rendering. Defaults to relying on the command 'vmd'.
        :param tachyon_executable: 'Path' to the tachyon executable to use for image rendering. Defaults to relying on the command 'tachyon'.
        :param vmd_logging: Whether to print output from vmd.
        """
        super().__init__(
            *args,
            cube_file = cube_file,
            rotations = rotations,
            auto_crop = auto_crop,
            resolution = resolution,
            also_make_png = also_make_png,
            isovalue = isovalue,
            num_cpu = num_cpu,
            **kwargs
        )
        
        # Some options that control how we function.
        self.rendering_style = rendering_style
        
        # Save executable paths.
        self.vmd_executable = vmd_executable
        self.tachyon_executable = tachyon_executable
        
        self.vmd_logging = vmd_logging
        
        
        if self.rendering_style is None:
            # Panic time.
            raise ValueError("rendering_style must not be None")
        
        # The colours we will be using for rendering (depends on rendering_stype and is only for reference; the actual definition of each style is in VMD).
        if self.rendering_style == "gaussian":
            self.primary_colour = "green"
            self.secondary_colour = "red"
        elif self.rendering_style == "vesta":
            self.primary_colour = "blue"
            self.secondary_colour = "yellow"
        else:
            self.primary_colour = "blue"
            self.secondary_colour = "red"
        
    @property
    def rotations(self):
        # Digichem rotates the wrong way round for some reason, reverse for our rendering engine.
        # VMD also likes degrees not radians.
        return [(axis, math.degrees(-theta)) for axis, theta in self._rotations]
    
    @classmethod
    def from_options(self, output, *, cube_file = None, rotations = None, options, num_cpu = 1, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            cube_file = cube_file,
            rotations = rotations,
            auto_crop = options['render']['auto_crop'],
            rendering_style = options['render']['vmd']['rendering_style'],
            resolution = options['render']['resolution'],
            isovalue = options['render'][self.options_name]['isovalue'],
            use_existing = options['render']['use_existing'],
            dont_modify = not options['render']['enable_rendering'],
            vmd_executable = options['render']['vmd']['executable'],
            tachyon_executable = options['render']['vmd']['tachyon'],
            vmd_logging = options['logging']['render_logging'],
            num_cpu = num_cpu,
            **kwargs
        )
    
    @classmethod
    def safe_name(self, file_name):
        """
        Get a filename free of 'unusual' characters.
        :return: The safe path name.
        """
        safe_chars = "._"
        return "".join([char if char.isalnum() or char in safe_chars else "_" for char in file_name])
    
    @property
    def vmd_script_path(self):
        """
        Get the file path to the VMD script to be used by this class to render images (as a pathlib Path).
        """
        return get_resource('data/vmd/{}'.format(self.vmd_script))
    
    @property
    def tcl_common_path(self):
        """
        Get the file path to the Tcl script that is common to all of our other Tcl scripts (as a pathlib Path).
        """
        return get_resource('data/vmd/common.tcl')
    
    @property
    def creation_message(self):
        """
        A short message that may (depending on log-level) be printed to the user before make_files() is called.
        """
        return "Rendering '{}' to file(s)".format(self.output if self.full_path_names else self.output.name)
        
    def make_files(self):
        """
        Make the image files referenced by this object.
        
        The new image will be written to file.
        """
        # Run VMD, which writes a plain text description of our scene to file.
        self.run_VMD_script()
        
        # Next run tachyon for each of our 4 scene files, which creates tga files.
        for image_name in ['x0y0z0', 'x90y0z0', 'x0y90z0', 'x45y45z45']:
            image_path = self.file_path[image_name]
            try:
                # First we'll render a test image at a lower resolution. We'll then crop it, and use the decrease in final resolution to know how much bigger we need to render in our final image to hit our target resolution.
                # Unless of course auto_crop is False, in which case we use our target resolution immediately.
                resolution = self.test_resolution if self.auto_crop else self.target_resolution
                self.run_tachyon_renderer(
                    image_path.with_name(self.safe_name(image_path.with_suffix(self.scene_file_extension).name)),
                    image_path.with_suffix(".tga"),
                    resolution
                )
                
                if self.auto_crop:
                    # Load the test image and autocrop it.
                    with Image.open(image_path.with_suffix(".tga"), "r") as test_im:
                        small_test_im = self.auto_crop_image(test_im)
                        
                    # Get the cropped size. We're interested in the largest dimension, as this is what we'll output as.
                    cropped_resolution = max(small_test_im.size)
                    
                    # From this we can work out the ratio between our true resolution and the resolution we've been asked for.
                    resolution_ratio = cropped_resolution / self.test_resolution
                    
                    # Now we can re-render asking for a larger resolution so that when we crop again, we should get our target resolution.
                    # Note that resolution ratio is a fraction, so self.target_resolution / resolution_ratio results in an increase.
                    self.run_tachyon_renderer(
                        image_path.with_name(self.safe_name(image_path.with_suffix(self.scene_file_extension).name)),
                        image_path.with_suffix(".tga"),
                        self.target_resolution / resolution_ratio
                    )
                
                # Delete the now unneeded scene file.
                os.remove(image_path.with_name(self.safe_name(image_path.with_suffix(self.scene_file_extension).name)))
            except Exception as e:
                raise File_maker_exception(self, "Error in tachyon rendering") from e
            
            # Convert to a better set of formats.
            # Open the file we just rendered.
            with Image.open(image_path.with_suffix(".tga"), "r") as im:
                
                # If we've been asked to autocrop, do so.
                if self.auto_crop:
                    try:
                        cropped_image = self.auto_crop_image(im)
                    except Exception:
                        raise File_maker_exception(self, "Error in post-rendering auto-crop")
                else:
                    cropped_image = im
                
                # Now save in our main output format.
                cropped_image.save(image_path)
                
                # And also as a higher quality png if we've been asked to.
                if self.also_make_png:
                    cropped_image.save(self.file_path[image_name + "_big"])
                
                # And delete the old .tga (unless we've just rendered to .tga!).
                if image_path.suffix != ".tga":
                    os.remove(image_path.with_suffix(".tga"))
        
    @property
    def prepared_translations(self):
        """
        Our list of translations in a from ready for VMD/Tcl.
        """
        return "{},{},{}".format(self.translations[0], self.translations[1], self.translations[2])
    
    @property
    def prepared_rotations(self):
        """
        Our list of rotations in a form ready for VMD/Tcl.
        """
        # This is a stringified form of our list of tuples that we'll pass to VMD (or rather the TCL interpreter).
        if len(self.rotations) > 0:
            rot_string = ":".join(["{},{}".format(axis, angle) for axis, angle in self.rotations])
        else:
            # An empty string will be misinterpreted on the command line.
            rot_string = "0,0"
        return rot_string
    
    @property
    def VMD_signature(self):
        """
        The signature pass to subprocess.run used to call VMD. Inheriting classes should write their own implementation.
        """
        return ""
    
    @property
    def inputs(self):
        """
        A dictionary of all the input files required by VMD for this image.
        
        inputs is a dictionary where each key is the path to the locally accessible file,
        and each value is the true location.
        """
        working_directory = self.output.parent
        return {
            Path(working_directory, self.vmd_script_path.name): self.vmd_script_path,
            Path(working_directory, self.tcl_common_path.name): self.tcl_common_path,
            Path(working_directory, self.safe_name(Path(str(self.input_file)).name)): Path(str(self.input_file)).absolute()
        }
    
    def run_VMD_script(self):
        """
        Called as part of the make() method, inheriting classes can implement this method if they have a different VMD call signature.
        
        Also see VMD_signature, which may do what you need.
        
        :return: The CompletedProcess object. 
        """
        # Setting these variables greatly improves VMD startup time.
        os.environ["VMDNOOPTIX"] = "1"
        os.environ["VMDNOOSPRAY"] = "1"
        
        # VMD has two run scripts, one in C-shell (which is obviously garbage) and one in bash.
        # Sadly, C-Shell is often the default, and it's doesn't support filenames properly.
        # We'll use the usual workaround of changing the working directory and using relative file names.
        working_directory = self.output.parent
        
        try:
            # Create local links to our input files.
            for input_dst, input_src in self.inputs.items():
                # Don't symlink if the source is already in the output dir.
                if input_dst.absolute() != input_src.absolute():
                    try:
                        os.symlink(input_src, input_dst)
                    
                    except Exception:
                        # Couldn't symlink for some reason, print a warning and copy instead.
                        digichem.log.get_logger().warning("Failed to create symlink for VMD input file, falling back to copy (this may take a while)")
                        shutil.copy(input_src, input_dst)
            
            digichem.log.get_logger().debug(str(self.VMD_signature))
            # Run VMD, which renders our image for us.
            try:
                return subprocess.run(
                    self.VMD_signature,
                    # We normally capture and discard stdout (because VMD is VERY verbose), but if we're at a suitable log level, we'll print it.
                    # Nothing useful appears to be printed to stderr, so we'll treat it the same as stdout.,
                    stdin = subprocess.DEVNULL,
                    stdout = subprocess.DEVNULL if not self.vmd_logging else None,
                    stderr = subprocess.STDOUT,
                    universal_newlines = True,
                    cwd = working_directory,
                    # VMD has a tendency to sigsegv when closing with VMDNOOPTIX set to on (even tho everything is fine) so we can't check retval sadly.
                    #check = True
                )
            except FileNotFoundError:
                raise File_maker_exception(self, "Could not locate vmd executable '{}'".format(self.vmd_executable))
        
        finally:
            # Clean up.
            # Remove the copied inputs
            for input_dst, input_src in self.inputs.items():
                # Don't symlink if the source is already in the output dir.
                if input_dst.absolute() != input_src.absolute():
                    try:
                        input_dst.unlink()
                    
                    except Exception:
                        # Warnings are useful here, if we can't delete the files we probably failed to copy them in the first place.
                        digichem.log.get_logger().warning("Failed to delete VMD input file '{}'".format(input_dst), exc_info = True)
            
        
    def run_tachyon_renderer(self, scene_file, tga_file, resolution):
        """
        Called as part of the make() method.
        
        Take a molecule 'scene' file produced by VMD and render to an image with tachyon.
        
        :param scene_file: A VMD scene file to read in as input.
        :param tga_file: The filename to write to.
        :param resolution: The resolution to render with.
        """
        # Sadly, tachyon is old and decrepit and doesn't parse the output filename correctly (but weirdly the input file is fine).
        # To get around this, we'll use a semi-random output name that we know is safe, and change-directory to the output directory immediately before calling tachyon.
        # This way we can ensure there will be no nasty file-names for tachyon to get stuck on.
        working_directory = tga_file.parent
        tmpfile_name = ".tachyon_output_" + uuid4().hex + ".tga"
        tmpfile_full_path = Path(tga_file.parent, tmpfile_name)
        tachyon_process = None
        try:
            try:
                # Now we can run tachyon.
                tachyon_process = subprocess.run(
                    [
                        "{}".format(self.tachyon_executable),
                        scene_file.relative_to(working_directory),
                        "-aasamples", "12",
                        # Note: this can get capped in a SLURM context...
                        "-numthreads", "{}".format(self.num_cpu),
                        "-res", "{}".format(resolution), "{}".format(resolution),
                        "-o", tmpfile_name
                    ],
                    stdout = subprocess.PIPE if not self.vmd_logging else None,
                    stderr = subprocess.STDOUT,
                    universal_newlines = True,
                    check = True,
                    cwd = working_directory
                )
            except FileNotFoundError:
                # THINK this can only occur if the command itself cannot be found, but I might be wrong.
                raise File_maker_exception(self, "Could not locate tachyon executable '{}'".format(self.tachyon_executable))
            
            # Render complete (sadly we don't appear to have any ability to check for errors).
            # Now rename our tmpfile.
            #os.rename(tmpfile_full_path, tga_file)
            tmpfile_full_path.rename(tga_file)
        
        except Exception:
            # Something went wrong, remove our tmpfile.
            try:
                os.remove(tmpfile_full_path)
            except Exception:
                pass
            
            # If we didn't already show output, dump it now.
            if not self.vmd_logging and tachyon_process is not None:
                digichem.log.get_logger().error("Tachyon did not exit successfully, dumping output:\n{}".format(tachyon_process.stdout))
                
            raise
        

class Structure_image_maker(VMD_image_maker):
    """
    Class for creating structure images.
    """
    
    vmd_script ="generate_structure_images.tcl"
    
    @property
    def VMD_signature(self):
        """
        The arguments which we'll pass to VMD, inheriting classes can implement this method if they have a different VMD call signature.
        """
        return [
                "{}".format(self.vmd_executable),
                "-dispdev", "none",
                "-e", "{}".format(self.vmd_script_path.name),
                "-args",
                "{}".format(self.safe_name(Path(str(self.input_file)).name)),
                "{}".format(self.tcl_common_path.name),
                "{}".format(self.rendering_style),
                "{}".format(self.prepared_translations),
                "{}".format(self.prepared_rotations),
                "{}".format(self.safe_name(self.file_path['x0y0z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x90y0z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x0y90z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x45y45z45'].with_suffix(self.scene_file_extension).name))
            ]
        
        
class Orbital_image_maker(Structure_image_maker):
    """
    Class for creating orbital images.
    """
    
    vmd_script ="generate_orbital_images.tcl"
    
    @property
    def VMD_signature(self):
        """
        The arguments which we'll pass to VMD, inheriting classes can implement this method if they have a different VMD call signature.
        """
        return [
                "{}".format(self.vmd_executable),
                "-dispdev", "none",
                "-e", "{}".format(self.vmd_script_path.name),
                "-args",
                "{}".format(self.safe_name(Path(str(self.input_file)).name)),
                "{}".format(self.tcl_common_path.name),
                "{}".format(self.rendering_style),
                "{}".format(self.isovalue),
                "{}".format(self.prepared_translations),
                "{}".format(self.prepared_rotations),
                "{}".format(self.safe_name(self.file_path['x0y0z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x90y0z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x0y90z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x45y45z45'].with_suffix(self.scene_file_extension).name))
            ]
    

class Difference_density_image_maker(Orbital_image_maker):
    
    # Name of the section where we get some specific configs.
    options_name = "difference_density"


class Spin_density_image_maker(Orbital_image_maker):
    """
    Class for creating spin density images.
    """
        
    vmd_script ="generate_spin_images.tcl"
    
    # Name of the section where we get some specific configs.
    options_name = "spin"
    
    def __init__(self, *args, spin = "both", **kwargs):
        """
        Constructor for Spin_density_image_maker objects.
        
        See Orbital_image_maker for a complete constructor.
        :param spin: A string indicating which net-spins to render, either 'positive', 'negative' or 'both'.
        """
        super().__init__(*args, **kwargs)
        self.spin = spin
                
    @property
    def VMD_signature(self):
        """
        The arguments which we'll pass to VMD, inheriting classes can implement this method if they have a different VMD call signature.
        """
        return [
                "{}".format(self.vmd_executable),
                "-dispdev", "none",
                "-e", "{}".format(self.vmd_script_path.name),
                "-args",
                "{}".format(self.safe_name(Path(str(self.input_file)).name)),
                "{}".format(self.tcl_common_path.name),
                "{}".format(self.rendering_style),
                "{}".format(fabs(self.isovalue)),
                "{}".format(self.spin),
                "{}".format(self.prepared_translations),
                "{}".format(self.prepared_rotations),
                "{}".format(self.safe_name(self.file_path['x0y0z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x90y0z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x0y90z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x45y45z45'].with_suffix(self.scene_file_extension).name))
            ]
    
class Alpha_orbital_image_maker(Orbital_image_maker):
    pass
    #cubegen_type = "AMO"
    
    
class Beta_orbital_image_maker(Orbital_image_maker):
    pass
    #cubegen_type = "BMO"
    
class Density_image_maker(Orbital_image_maker):
    """
    Class for creating spin density images.
    """
        
    vmd_script ="generate_density_images.tcl"
    
    # Name of the section where we get some specific configs.
    options_name = "density"
    
    def __init__(self, *args, **kwargs):
        """
        Constructor for Spin_density_image_maker objects.
        
        See Orbital_image_maker for a complete constructor.
        """
        super().__init__(*args, **kwargs)
    
    @property
    def type(self):
        """
        The density type.
        """
        return self.input_file.type
                
    @property
    def VMD_signature(self):
        """
        The arguments which we'll pass to VMD, inheriting classes can implement this method if they have a different VMD call signature.
        """
        return [
                "{}".format(self.vmd_executable),
                "-dispdev", "none",
                "-e", "{}".format(self.vmd_script_path.name),
                "-args",
                "{}".format(self.safe_name(Path(str(self.input_file)).name)),
                "{}".format(self.tcl_common_path.name),
                "{}".format(self.rendering_style),
                "{}".format(fabs(self.isovalue)),
                "{}".format(self.prepared_translations),
                "{}".format(self.prepared_rotations),
                "{}".format(self.safe_name(self.file_path['x0y0z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x90y0z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x0y90z0'].with_suffix(self.scene_file_extension).name)),
                "{}".format(self.safe_name(self.file_path['x45y45z45'].with_suffix(self.scene_file_extension).name))
            ]
    
class Combined_orbital_image_maker(VMD_image_maker):
    """
    Class for creating images with both the HOMO and LUMO shown together.
    """
    
    vmd_script ="generate_combined_orbital_images.tcl"
    
    def __init__(self, *args, HOMO_cube_file = None, LUMO_cube_file = None, **kwargs):
        """
        Constructor for combined orbital image maker objects.
        
        :param output: Path to write to. See the constructor for VMD_image_maker for how this works.
        :param HOMO_cube_file: Path to the HOMO cube file.
        :param LUMO_cube_file: Path to the LUMO cube file.
        :param *args: See the constructor for VMD_image_maker for further options.
        :param **kwargs: See the constructor for VMD_image_maker for further options.
        """
        super().__init__(*args, cube_file = None, **kwargs)
        self.HOMO_cube_file = HOMO_cube_file
        self.LUMO_cube_file = LUMO_cube_file
        
    @classmethod
    def from_options(self, output, *, HOMO_cube_file = None, LUMO_cube_file = None, rotations = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            HOMO_cube_file = HOMO_cube_file,
            LUMO_cube_file = LUMO_cube_file,
            rotations = rotations,
            auto_crop = options['render']['auto_crop'],
            rendering_style = options['render']['vmd']['rendering_style'],
            resolution = options['render']['resolution'],
            isovalue = options['render'][self.options_name]['isovalue'],
            use_existing = options['render']['use_existing'],
            dont_modify = not options['render']['enable_rendering'],
            vmd_executable = options['render']['vmd']['executable'],
            tachyon_executable = options['render']['vmd']['tachyon'],
            vmd_logging = options['logging']['render_logging'],
            **kwargs
        )
    
    def check_can_make(self):
        """
        Check whether it is feasible to try and render the image(s) that we represent.
        
        Reasons for rendering not being possible are varied and are up to the inheriting class, but include eg, a required input (cube, fchk) file not being given.
        
        This method returns nothing, but will raise a File_maker_exception exception if the rendering is not possible.
        """
        try:
            if self.HOMO_cube_file is None or self.HOMO_cube_file.safe_get_file() is None:
                raise File_maker_exception(self, "No HOMO cube file is available")
        except AttributeError:
            if not hasattr(self.HOMO_cube_file, 'safe_get_file'):
            # Input file does not have a safe_get_file() method.
                pass

            else:
                raise

        try:
            if self.LUMO_cube_file is None  or self.LUMO_cube_file.safe_get_file() is None:
                raise File_maker_exception(self, "No LUMO cube file is available")
        
        except AttributeError:
            if not hasattr(self.LUMO_cube_file, 'safe_get_file'):
            # Input file does not have a safe_get_file() method.
                pass

            else:
                raise
    
    @property
    def inputs(self):
        """
        A dictionary of all the input files required by VMD for this image.
        
        inputs is a dictionary where each key is the path to the locally accessible file,
        and each value is the true location.
        """
        working_directory = self.output.parent
        return {
            Path(working_directory, self.vmd_script_path.name): self.vmd_script_path,
            Path(working_directory, self.tcl_common_path.name): self.tcl_common_path,
            Path(working_directory, self.safe_name(Path(str(self.HOMO_cube_file)).name)): Path(str(self.HOMO_cube_file)).absolute(),
            Path(working_directory, self.safe_name(Path(str(self.LUMO_cube_file)).name)): Path(str(self.LUMO_cube_file)).absolute()
        }
    
    @property
    def VMD_signature(self):
        """
        The arguments which we'll pass to VMD, inheriting classes can implement this method if they have a different VMD call signature.
        """
        return [
            "{}".format(self.vmd_executable),
            "-dispdev", "none",
            "-e", "{}".format(self.vmd_script_path.name),
            "-args",
            "{}".format(self.safe_name(Path(str(self.HOMO_cube_file)).name)),
            "{}".format(self.safe_name(Path(str(self.LUMO_cube_file)).name)),
            "{}".format(self.tcl_common_path.name),
            "{}".format(self.rendering_style),
            "{}".format(fabs(self.isovalue)),
            "{}".format(self.prepared_translations),
            "{}".format(self.prepared_rotations),
            "{}".format(self.safe_name(self.file_path['x0y0z0'].with_suffix(self.scene_file_extension).name)),
            "{}".format(self.safe_name(self.file_path['x90y0z0'].with_suffix(self.scene_file_extension).name)),
            "{}".format(self.safe_name(self.file_path['x0y90z0'].with_suffix(self.scene_file_extension).name)),
            "{}".format(self.safe_name(self.file_path['x45y45z45'].with_suffix(self.scene_file_extension).name))
        ]


    
class Dipole_image_maker(Structure_image_maker):
    """
    Class for creating dipole images.
    """
    
    vmd_script ="generate_dipole_images.tcl"
    
    
    def __init__(self, *args, dipole_moment = None, magnetic_dipole_moment = None, scaling = 1, magnetic_scaling = 1, **kwargs):
        """
        Constructor for Dipole_image_maker objects.
        
        :param output: Path to write to. See the constructor for VMD_image_maker for how this works.
        :param cube_file: A Gaussian cube file to use to render the new images.
        :param dipole_moment: A Dipole_moment object that will be rendered as a red arrow in the scene.
        :param magnetic_dipole_moment: A second dipole moment object that will be rendered as a blue arrow in the scene.
        :param scaling: A factor to scale the dipole moment by.
        :param magnetic_scaling: A factor to scale the magnetic dipole moment by.
        :param *args: See the constructor for VMD_image_maker for further options.
        :param **kwargs: See the constructor for VMD_image_maker for further options.
        """
        super().__init__(*args, **kwargs)
        self.dipole_moment = dipole_moment
        self.magnetic_dipole_moment = magnetic_dipole_moment
        self.scaling = scaling
        self.magnetic_scaling = magnetic_scaling
        self.electric_arrow_colour = "red"
        self.magnetic_arrow_colour = "green"

    def get_coords(self, dipole, scaling):
        if dipole is None:
            return ( (0.0,0.0,0.0), (0.0,0.0,0.0))
        
        else:
            return (
                tuple([coord * scaling for coord in dipole._origin_coords]),
                tuple([coord * scaling for coord in dipole._vector_coords])
            ) 
        
    @property
    def VMD_signature(self):
        """
        The arguments which we'll pass to VMD, inheriting classes can implement this method if they have a different VMD call signature.
        """
        return [
            "{}".format(self.vmd_executable),
            "-dispdev", "none",
            "-e", "{}".format(self.vmd_script_path.name),
            "-args",
            "{}".format(self.safe_name(Path(str(self.input_file)).name)),
            "{}".format(self.tcl_common_path.name),
            "{}".format(self.rendering_style),
            "{}".format(self.prepared_translations),
            "{}".format(self.prepared_rotations),
            # We don't use the normal origin_coords/vector_coords because these are already rotated, while we want/need to do this rotation with the camera in VMD.
            # Hence use _origin_coords/_vector_coods, which aren't rotated.
            # Dipole 1 (electric).
            "{}:{}:{}".format(*self.get_coords(self.dipole_moment, self.scaling)[0]),
            "{}:{}:{}".format(*self.get_coords(self.dipole_moment, self.scaling)[1]),
            # Dipole 2 (magnetic).
            "{}:{}:{}".format(*self.get_coords(self.magnetic_dipole_moment, self.magnetic_scaling)[0]),
            "{}:{}:{}".format(*self.get_coords(self.magnetic_dipole_moment, self.magnetic_scaling)[1]),
            "{}".format(self.safe_name(self.file_path['x0y0z0'].with_suffix(self.scene_file_extension).name)),
            "{}".format(self.safe_name(self.file_path['x90y0z0'].with_suffix(self.scene_file_extension).name)),
            "{}".format(self.safe_name(self.file_path['x0y90z0'].with_suffix(self.scene_file_extension).name)),
            "{}".format(self.safe_name(self.file_path['x45y45z45'].with_suffix(self.scene_file_extension).name))
        ]

    
class Permanent_dipole_image_maker(Dipole_image_maker):
    """
    Class for creating dipole images.
    """
    
    def __init__(self, *args, dipole_moment = None, scaling = 1, **kwargs):
        """
        Constructor for Dipole_image_maker objects.
        
        :param output: Path to write to. See the constructor for VMD_image_maker for how this works.
        :param cube_file: A Gaussian cube file to use to render the new images.
        :param dipole_moment: A Dipole_moment object that will be rendered as a red arrow in the scene.
        :param scaling: A factor to scale the dipole moment by.
        :param *args: See the constructor for VMD_image_maker for further options.
        :param **kwargs: See the constructor for VMD_image_maker for further options.
        """
        super().__init__(*args, dipole_moment = dipole_moment, scaling = scaling, **kwargs)
        
    @classmethod
    def from_options(self, output, *, dipole_moment = None, cube_file = None, rotations = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            cube_file = cube_file,
            rotations = rotations,
            dipole_moment = dipole_moment,
            auto_crop = options['render']['auto_crop'],
            rendering_style = options['render']['vmd']['rendering_style'],
            resolution = options['render']['resolution'],
            isovalue = options['render'][self.options_name]['isovalue'],
            use_existing = options['render']['use_existing'],
            dont_modify = not options['render']['enable_rendering'],
            vmd_executable = options['render']['vmd']['executable'],
            tachyon_executable = options['render']['vmd']['tachyon'],
            vmd_logging = options['logging']['render_logging'],
            scaling = options['render']['dipole_moment']['scaling'],
            **kwargs
        )
    
    def check_can_make(self):
        """
        Check whether it is feasible to try and render the image(s) that we represent.
        
        Reasons for rendering not being possible are varied and are up to the inheriting class, but include eg, a required input (cube, fchk) file not being given.
        
        This method returns nothing, but will raise a File_maker_exception exception if the rendering is not possible.
        """ 
        super().check_can_make()
        
        # Also make sure we have a dipole.
        if self.dipole_moment is None:
            raise File_maker_exception(self, "No dipole moment is available.")
    

class Transition_dipole_image_maker(Dipole_image_maker):
    """
    Class for creating TDM images.
    """
    
    def check_can_make(self):
        """
        Check whether it is feasible to try and render the image(s) that we represent.
        
        Reasons for rendering not being possible are varied and are up to the inheriting class, but include eg, a required input (cube, fchk) file not being given.
        
        This method returns nothing, but will raise a File_maker_exception exception if the rendering is not possible.
        """ 
        super().check_can_make()
        
        # Also make sure we have a dipole.
        if self.dipole_moment is None and self.magnetic_dipole_moment is None:
            raise File_maker_exception(self, "No dipole moment is available.")
        
    @classmethod
    def from_options(self, output, *, dipole_moment = None, magnetic_dipole_moment, cube_file = None, rotations = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            cube_file = cube_file,
            rotations = rotations,
            dipole_moment = dipole_moment,
            magnetic_dipole_moment = magnetic_dipole_moment,
            auto_crop = options['render']['auto_crop'],
            rendering_style = options['render']['vmd']['rendering_style'],
            resolution = options['render']['resolution'],
            isovalue = options['render'][self.options_name]['isovalue'],
            use_existing = options['render']['use_existing'],
            dont_modify = not options['render']['enable_rendering'],
            vmd_executable = options['render']['vmd']['executable'],
            tachyon_executable = options['render']['vmd']['tachyon'],
            vmd_logging = options['logging']['render_logging'],
            scaling = options['render']['transition_dipole_moment']['electric_scaling'],
            magnetic_scaling = options['render']['transition_dipole_moment']['magnetic_scaling'],
            **kwargs
        )
    


import deepmerge

# Hidden imports.
# import basis_set_exchange.misc
# import basis_set_exchange.writers

class BSE_basis_set(dict):
    """
    A class for representing a (number of) basis sets fetched from the basis set exchange.
    """
    
    def __init__(self, *args, **kwargs):
        """
        Constructor for BSE_basis_set objects.
        
        :param definition: A dictionary where each key is the name of a basis set and the value is the elements it applies to. The values 'all', '*', and None can be used to indicate the basis set applies to all atoms.
        """
        super().__init__( *args, **kwargs)
        
    @classmethod
    def is_filter_all(self, element_filter):
        """
        Determine whether a given element filter implicitly includes all elements.
        """
        return element_filter == "*" or element_filter == "all" or element_filter is None
    
    def __str__(self):
        return self.name
    
    @property
    def name(self):
        """
        Get a descriptive name of this basis set.
        """
        import basis_set_exchange.misc
        
        names = []
        
        for basis_set_name, basis_set_elements in self.items():
            name = basis_set_name
            if not self.is_filter_all(basis_set_elements):
                name += " []".format(basis_set_exchange.misc.compact_elements(basis_set_exchange.misc.expand_elements(basis_set_elements)))
                
        return ", ".join(names)
    
    def has_ECPs(self, elements_filter = None):
        """
        Determine whether this basis set would provide effective core potentials (ECPs) when applied to a list of elements.
        
        :param elements_filter: The elements to check against.
        """
        basis_sets = self.to_dict(elements_filter)
        
        return 'scalar_ecp' in basis_sets['function_types']
    
    def to_dict(self, elements_filter = None):
        """
        Convert the basis set information represented by this object to a basis set dict.
        
        :param elements_filter: A list/string of elements to filter by. Only elements given here will be printed in the final format. Each item can be an int or str representing a single element (1, '1', 'H' etc), or a range of elements ('1-5' etc).
        """
        import basis_set_exchange.misc
        
        if elements_filter is not None and not isinstance(elements_filter, str):
            elements_filter = ",".join((str(item) for item in elements_filter))

        if elements_filter is not None:
            elements_filter = basis_set_exchange.misc.expand_elements(elements_filter)
        
        basis_sets = {}
        
        # First, convert each definition to to a basis_set_exchange dict
        for basis_set_name, basis_set_elements in self.items():
            # Only elements that are actually present will have a basis set recorded.
            # The values of 'all', '*' and None are also accepted as meaning all atoms.
            if not self.is_filter_all(basis_set_elements):
                basis_set_elements = basis_set_exchange.misc.expand_elements(basis_set_elements)
                if elements_filter is not None:
                    elements = list(set(basis_set_elements).intersection(elements_filter))
                    
                else:
                    elements = basis_set_elements
            
            else:
                # We are using all elements from this basis set.
                elements = elements_filter
            
            if elements is None or len(elements) > 0:
                # Get the basis set.
                basis_set = basis_set_exchange.get_basis(basis_set_name, elements)
                
                # Merge it with our total basis set.
                # TODO: Check this is actually safe?
                deepmerge.always_merger.merge(basis_sets, basis_set)
            
        return basis_sets
    
    def to_format(self, fmt = None, elements_filter = None):
        """
        Convert the basis set information represented by this object to a basis set format (for example, one suitable for a CC program).
        
        :param fmt: The format to write to, see basis_set_exchange.get_formats().
        :param elements_filter: A list of elements to filter by. Only elements given here will be printed in the final format. Each item can be an int or str representing a single element (1, '1', 'H' etc), or a range of elements ('1-5' etc).
        :returns: The formatted basis set.
        """
        import basis_set_exchange.writers
        
        basis_sets = self.to_dict(elements_filter)
            
        # Output the merged basis set.
        if fmt is not None:
            return basis_set_exchange.writers.write_formatted_basis_str(basis_sets, fmt = fmt)
        
        else:
            return basis_sets
    